#!/usr/bin/env python

"""
usage: schemalib.py [-h] [--indent INDENT] [-v] infilename

synopsis:
  parse the complexTypes in an XML schema.  Create objects.  Display them.

positional arguments:
  infilename       a reference to the input XML schema

optional arguments:
  -h, --help       show this help message and exit
  --indent INDENT  indentation. Default is 4 spaces
  -v, --verbose    Print messages during actions.

examples:
  python schemalib.py my_schema.xsd
  python schemalib.py -v my_schema.xsd

license:
  Copyright 2017 Dave Kuhlman

  Redistribution and use in source and binary forms, with or
  without modification, are permitted provided that the following
  conditions are met:

  1. Redistributions of source code must retain the above
  copyright notice, this list of conditions and the following
  disclaimer.

  2. Redistributions in binary form must reproduce the above
  copyright notice, this list of conditions and the following
  disclaimer in the documentation and/or other materials provided with
  the distribution.

  3. Neither the name of the copyright holder nor the names of its
  contributors may be used to endorse or promote products derived from
  this software without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
  IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
  ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""


#
# imports
from __future__ import print_function
import sys
import os
import argparse
from lxml import etree
import re


#
# Global variables

Opts = None
Ns_prefix_pat = re.compile(r'^{.*}')


#
# Classes

#
# Xsd/schema Classes

class XsdElementClass(object):
    """Representation of a top-level element declaration."""

    def __init__(self, node, nsmap=None):
        """Build an XsdElementClass instance given an Lxml node."""
        self.node = node
        self.name = None
        self.type_name = None
        self.type_obj = None
        if nsmap is None:
            self.nsmap = node.nsmap
        else:
            self.nsmap = nsmap
        self.name, self.type_name = get_element_name_type(self.node)
        if self.type_name is None:
            # look for an anonymous complexType definition.
            type_nodes = node.xpath(
                './xs:complexType',
                namespaces=self.nsmap)
            if type_nodes:
                type_node = type_nodes[0]
                self.type_obj = XsdComplexTypeClass(type_node, self.nsmap)
        dbg_msg("Created Element: {}".format(self))

    def show(self, wrt=sys.stdout.write, indent=""):
        """Display (print out) an Element."""
        if self.type_name:
            wrt('{}Element name: {:>20s}  type: {}\n'.format(
                indent, self.name, self.type_name))
        elif self.type_obj:
            wrt('{}element -- name: {} -- Type:\n'.format(indent, self.name))
            indent += '    '
            self.type_obj.show(wrt, indent)
        else:
            wrt('{}[no type defined\n'.format(indent))

    def __str__(self):
        return '<XsdElementClass -- name: {}>'.format(self.name)
    __repr__ = __str__
# end class XsdElementClass


class XsdComplexTypeClass(object):
    """Representation of an xs:complexType element definition."""

    def __init__(self, node, nsmap=None):
        """Build an XsdComplexTypeClass instance given an Lxml node."""
        self.node = node
        self.name = None
        self.sequence = []
        self.extension_base = None
        self.complex_content_sequence = []
        self.attributes = []
        self.attribute_groups = []
        if nsmap is None:
            self.nsmap = node.nsmap
        else:
            self.nsmap = nsmap
        self.name, _ = get_element_name_type(self.node)
        member_nodes = node.xpath(
            './xs:sequence/xs:element',
            namespaces=self.nsmap)
        for member_node in member_nodes:
            member = XsdMemberClass(member_node, self.nsmap)
            self.sequence.append(member)
        if not self.sequence:
            # try xs:complexContent.
            complex_content_nodes = node.xpath(
                './xs:complexContent',
                namespaces=self.nsmap)
            if complex_content_nodes:
                complex_content_node = complex_content_nodes[0]
                extension_nodes = complex_content_node.xpath(
                    './xs:extension',
                    namespaces=self.nsmap)
                if extension_nodes:
                    extension_node = extension_nodes[0]
                    self.extension_base = extension_node.get('base')
                    member_nodes = extension_node.xpath(
                        './xs:sequence/xs:element',
                        namespaces=self.nsmap)
                    for member_node in member_nodes:
                        member = XsdMemberClass(member_node, self.nsmap)
                        self.complex_content_sequence.append(member)
        attribute_nodes = node.xpath(
            './xs:attribute',
            namespaces=self.nsmap)
        for attribute_node in attribute_nodes:
            attribute = XsdAttributeClass(attribute_node, nsmap)
            self.attributes.append(attribute)
        attribute_group_nodes = node.xpath(
            './xs:attributeGroup',
            namespaces=self.nsmap)
        for attribute_group_node in attribute_group_nodes:
            attribute_group = XsdAttributeGroupClass(
                attribute_group_node, nsmap)
            self.attribute_groups.append(attribute_group)
        dbg_msg("Created ComplexType: {}".format(self))

    def show(self, wrt=sys.stdout.write, indent=""):
        """Display (print out) a ComplexType."""
        wrt('{}complexType -- name: {}\n'.format(indent, self.name))
        wrt('{}sequence:\n'.format(indent))
        indent += '    '
        for member in self.sequence:
            member.show(wrt, indent)
        if self.extension_base:
            wrt('{}extension base: {}\n'.format(indent, self.extension_base))
            for member in self.complex_content_sequence:
                member.show(wrt, indent)
        for attribute in self.attributes:
            attribute.show(wrt, indent)
        for attribute_group in self.attribute_groups:
            attribute_group.show(wrt, indent)

    def __str__(self):
        return '<XsdComplexTypeClass -- name: {}>'.format(self.name)
    __repr__ = __str__
# end class XsdComplexTypeClass


class XsdSimpleTypeClass(object):
    """Representation of an xs:simpleType element definition."""

    def __init__(self, node, nsmap=None):
        """Build an XsdSimpleType Class instance given an Lxml node."""
        self.node = node
        self.name = None
        self.restrictions = []
        self.lists = []
        self.unions = []
        if nsmap is None:
            self.nsmap = node.nsmap
        else:
            self.nsmap = nsmap
        self.name = node.get('name')
        restriction_nodes = node.xpath(
            './xs:restriction',
            namespaces=self.nsmap)
        if restriction_nodes:
            restriction_node = restriction_nodes[0]
            self.base = restriction_node.get('base')
            qualifier_nodes = restriction_node.xpath(
                './xs:minInclusive | '
                './xs:maxInclusive | '
                './xs:minExclusive | '
                './xs:maxExclusive | '
                './xs:enumeration | '
                './xs:minLength | '
                './xs:maxLength',
                namespaces=self.nsmap)
            for qualifier_node in qualifier_nodes:
                tag = strip_ns_prefix(qualifier_node.tag)
                value = qualifier_node.get('value')
                self.restrictions.append((tag, value))
        list_nodes = node.xpath(
            './xs:list',
            namespaces=self.nsmap)
        if list_nodes:
            for list_node in list_nodes:
                self.lists.append(list_node.get('itemType'))
        union_nodes = node.xpath(
            './xs:union',
            namespaces=self.nsmap)
        if union_nodes:
            for union_node in union_nodes:
                self.unions.append(union_node.get('memberTypes'))
        dbg_msg("Created SimpleType: {}".format(self))

    def show(self, wrt=sys.stdout.write, indent=""):
        """Display (print out) a SimpleType."""
        wrt('{}simpleType -- name: {}\n'.format(indent, self.name, ))
        for restriction, value in self.restrictions:
            wrt('    {}restriction: {:>20s}  value: {}\n'.format(
                indent, restriction, value))
        for listval in self.lists:
            wrt('    {}list: {}\n'.format(indent, listval))
        for unionval in self.unions:
            wrt('    {}union: "{}"\n'.format(indent, unionval))

    def __str__(self):
        return '<XsdSimpleTypeClass -- name: {}>'.format(self.name)
    __repr__ = __str__
# end class XsdSimpleTypeClass


class XsdMemberClass(object):
    """Representation of a member in a child member declaration."""

    def __init__(self, node, nsmap=None):
        """Build an XsdMemberClass instance given an Lxml node."""
        self.node = node
        self.name = None
        self.type_name = None
        if nsmap is None:
            self.nsmap = node.nsmap
        else:
            self.nsmap = nsmap
        self.name, self.type_name = get_element_name_type(node)
        if self.type_name is None:
            # try to find a nested anonymous complexType.
            type_nodes = node.xpath(
                './xs:complexType',
                namespaces=self.nsmap)
            if type_nodes:
                type_node = type_nodes[0]
                self.type_obj = XsdComplexTypeClass(type_node, self.nsmap)
        dbg_msg("Created Member: {}".format(self))

    def show(self, wrt=sys.stdout.write, indent=""):
        """Display (print out) a Member/child."""
        wrt('{}name: {:>20s}  type: {}\n'.format(
            indent, self.name, self.type_name))

    def __str__(self):
        return '<XsdMemberClass -- name: {} type_name: {}>'.format(
            self.name, self.type_name)
    __repr__ = __str__
# end class XsdMemberClass


class XsdAttributeGroupClass(object):
    """Representation of an xs:attributeGroup element definition."""

    def __init__(self, node, nsmap=None):
        """Build an XsdAttributeGroupClass instance given an Lxml node."""
        self.node = node
        self.name = None
        self.type_name = None
        self.attributes = []
        if nsmap is None:
            self.nsmap = node.nsmap
        else:
            self.nsmap = nsmap
        self.name, self.type_name = get_element_name_type(node)
        attribute_nodes = node.xpath(
            './xs:attribute',
            namespaces=self.nsmap)
        for attribute_node in attribute_nodes:
            attribute = XsdAttributeClass(attribute_node)
            self.attributes.append(attribute)
        dbg_msg("Created AttributeGroup: {}".format(self))

    def show(self, wrt=sys.stdout.write, indent=""):
        """Display (print out) an AttributeGroup."""
        wrt('{}attributeGroup  -- name: {}\n'.format(indent, self.name, ))
        indent += '    '
        for attribute in self.attributes:
            attribute.show(wrt, indent)

    def __str__(self):
        return '<XsdAttributeGroupClass -- name: {}>'.format(self.name)
    __repr__ = __str__
# end class XsdAttributeGroupClass


class XsdAttributeClass(object):
    """Representation of an xs:attribute element definition."""

    def __init__(self, node, nsmap=None):
        """Build an XsdAttributeClass instance given an Lxml node."""
        self.node = node
        self.name = None
        self.type_name = None
        if nsmap is None:
            self.nsmap = node.nsmap
        else:
            self.nsmap = nsmap
        self.name = node.get('name')
        self.name, self.type_name = get_element_name_type(node)
        dbg_msg("Created Attribute: {}".format(self))

    def show(self, wrt=sys.stdout.write, indent=""):
        """Display (print out) an Attribute."""
        wrt('{}attribute -- name: {}  type: {}\n'.format(
            indent, self.name, self.type_name, ))

    def __str__(self):
        return '<XsdAttributeClass -- name: {}>'.format(self.name)
    __repr__ = __str__
# end class XsdAttributeClass


class XsdSchemaClass(object):
    """Container and utility methods for dealing with an XML schema doc."""

    def __init__(self):
        """Initialize the member variables in a schema representation."""
        self.location = None
        self.nsmap = None
        self.doc = None
        self.root = None
        self.complexType_nodes = None
        self.simpleType_nodes = None
        self.element_nodes = None
        self.attributeGroup_nodes = None
        self.complex_types = None
        self.simple_types = None
        self.elements = None
        self.attribute_groups = None
        self.complex_type_map = None
        self.simple_type_map = None
        self.element_map = None
        self.attribute_group_map = None
        dbg_msg("Created Schema: {}".format(self))

    def parse(self, ref):
        """Parse a schema; return (doc,root)."""
        self.location = os.path.abspath(ref)
        self.doc = etree.parse(ref)
        self.root = self.doc.getroot()
        self.nsmap = self.root.nsmap.copy()
        if None in self.nsmap:
            del self.nsmap[None]
        return self.doc, self.root

    def build_schema(self):
        """Retrieve and build top-level element, complexType, and simpleType"""
        self.complex_type_nodes = self.root.xpath(
            './xs:complexType',
            namespaces=self.nsmap)
        self.simple_type_nodes = self.root.xpath(
            './xs:simpleType',
            namespaces=self.nsmap)
        self.element_nodes = self.root.xpath(
            './xs:element',
            namespaces=self.nsmap)
        self.attributeGroup_nodes = self.root.xpath(
            './xs:attributeGroup',
            namespaces=self.nsmap)
        self.complex_types = [
            XsdComplexTypeClass(ctnode, self.nsmap)
            for ctnode in self.complex_type_nodes]
        self.simple_types = [
            XsdSimpleTypeClass(stnode, self.nsmap)
            for stnode in self.simple_type_nodes]
        self.elements = [
            XsdElementClass(elnode, self.nsmap)
            for elnode in self.element_nodes]
        self.attribute_groups = [
            XsdAttributeGroupClass(elnode, self.nsmap)
            for elnode in self.attributeGroup_nodes]
        self.complex_type_map = {obj.name: obj for obj in self.complex_types}
        self.simple_type_map = {obj.name: obj for obj in self.simple_types}
        self.attribute_group_map = {
            obj.name: obj for obj in self.attribute_groups}
        self.element_map = {obj.name: obj for obj in self.elements}

    def print_schema(self, outfile):
        """Display (print or write) the schema and the objects in it."""
        wrt = outfile.write
        wrt('{}\n'.format('=' * 60))
        wrt('    Complex types\n')
        for index, ct in enumerate(self.complex_types):
            wrt('----- {}. {}\n'.format(index + 1, '-' * 30))
            ct.show(wrt)
        wrt('{}\n'.format('=' * 60))
        wrt('    Simple types\n')
        for index, st in enumerate(self.simple_types):
            wrt('----- {}. {}\n'.format(index + 1, '-' * 30))
            st.show(wrt)
        wrt('{}\n'.format('=' * 60))
        wrt('    Element declarations\n')
        for index, el in enumerate(self.elements):
            wrt('----- {}. {}\n'.format(index + 1, '-' * 30))
            el.show(wrt)
        wrt('{}\n'.format('=' * 60))
        wrt('    AttributeGroup declarations\n')
        for index, attr_group in enumerate(self.attribute_groups):
            wrt('----- {}. {}\n'.format(index + 1, '-' * 30))
            attr_group.show(wrt)

    def __str__(self):
        return '<XsdSchemaClass -- location: {}>'.format(self.location)
    __repr__ = __str__
# end class XsdSchemaClass


#
# Xsd/schema Classes

class XmlDoc(object):
    """A representation of an XML instance document."""

    def __init__(self, schema_ref, ref):
        """Build the schema objects, then build the instance objects.

        :param schema_ref: The schema.  Maybe anything accepted by
            lxml.etree.parse().
        :type schema_ref: str.
        :param ref: The XML instance doc.  Maybe anything accepted by
            lxml.etree.parse().
        :type ref: str.
        """
        self.instance = None
        self.schema = XsdSchemaClass()
        self.schema.parse(schema_ref)
        self.schema.build_schema()
        # create the instance (self.instance).
        self.instance = self.build_instance(ref)

    def build_instance(self, ref):
        schema = self.schema
        instance = None
        doc = etree.parse(ref)
        root = doc.getroot()
        print('schema: {}  root:{}'.format(schema, root))
        return instance


class XmlElement(object):
    pass


#
# Exported functions

def test_schemalib(infilename, outfilename):
    """Run a test.

    :param infilename: The name of the input XML schema file.
    :type infilename: str.
    :param outfilename: The name of the output XML schema file.
    :type outfilename: str or None.
    """
    if outfilename:
        outfile = make_file(outfilename)
    else:
        outfile = sys.stdout
    schema = XsdSchemaClass()
    schema.parse(infilename)
    schema.build_schema()
    schema.print_schema(outfile)
    if outfilename:
        outfile.close()


def set_global_opts(verbose=False, force=False):
    """Intialize global variable Opts to needed options."""
    global Opts
    opts = {
        'verbose': verbose,
        'force': force,
    }
    Opts = opts


#
# Private functions

def get_element_name_type(node):
    name = node.attrib.get('ref')
    if name:
        type_name = name
    else:
        name = node.attrib.get('name')
        type_name = node.attrib.get('type')
    return name, type_name


def strip_ns_prefix(tag):
    bare_tag = re.sub(Ns_prefix_pat, '', tag)
    return bare_tag


def make_file(outfilename):
    if os.path.exists(outfilename) and not Opts['force']:
        sys.exit("outfilename exists.  Use --force to overwrite.")
    outfile = open(outfilename, 'w')
    return outfile


def dbg_msg(msg):
    """Print a message if verbose is on."""
    if Opts['verbose']:
        print(msg)


def main():
    global Opts
    description = """\
synopsis:
  parse the complexTypes in an XML schema.  Create objects.  Display them.
"""
    epilog = """\
examples:
  python schemalib.py my_schema.xsd
  python schemalib.py -v my_schema.xsd
"""
    parser = argparse.ArgumentParser(
        description=description,
        epilog=epilog,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    parser.add_argument(
        "infilename",
        help="a reference to the input XML schema"
    )
    parser.add_argument(
        "-o", "--outfilename",
        type=str, dest="outfilename",
        action="store", default=None,
        help="indentation.  Default is 4 spaces",
    )
    parser.add_argument(
        "--indent",
        help="indentation.  Default is 4 spaces"
    )
    parser.add_argument(
        "-f", "--force",
        action="store_true",
        help="overwrite output file",
    )
    parser.add_argument(
        "-v", "--verbose",
        action="store_true",
        help="Print messages during actions.",
    )
    options = parser.parse_args()
    options = vars(options)
    Opts = options
    dbg_msg('options: {}'.format(options))
    infilename = options['infilename']
    outfilename = options['outfilename']
    test_schemalib(infilename, outfilename)


if __name__ == '__main__':
    #import pdb; pdb.set_trace()
    #import ipdb; ipdb.set_trace()
    # or, at the IPython interactive prompt, use `%run -d ...`.
    main()
